# -*- mode: python -*-
# -*- coding: utf-8 -*-

##############################################################################
#
# ScoDoc
#
# Copyright (c) 1999 - 2022 Emmanuel Viennet.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Emmanuel Viennet      emmanuel.viennet@viennet.net
#
##############################################################################

"""
Formulaires configuration Exports Apogée (codes)
"""

from flask import flash, url_for, redirect, request, render_template
from flask_wtf import FlaskForm
from wtforms import BooleanField, SelectField, SubmitField

import app
from app.models import ScoDocSiteConfig
import app.scodoc.sco_utils as scu


class BonusConfigurationForm(FlaskForm):
    "Panneau de configuration des logos"
    bonus_sport_func_name = SelectField(
        label="Fonction de calcul des bonus sport&culture",
        choices=[
            (name, displayed_name if name else "Aucune")
            for (name, displayed_name) in ScoDocSiteConfig.get_bonus_sport_class_list()
        ],
    )
    submit_bonus = SubmitField("Valider")
    cancel_bonus = SubmitField("Annuler", render_kw={"formnovalidate": True})


class ScoDocConfigurationForm(FlaskForm):
    "Panneau de configuration avancée"
    enable_entreprises = BooleanField("activer le module <em>entreprises</em>")
    submit_scodoc = SubmitField("Valider")
    cancel_scodoc = SubmitField("Annuler", render_kw={"formnovalidate": True})


def configuration():
    "Page de configuration principale"
    # nb: le contrôle d'accès (SuperAdmin) doit être fait dans la vue
    form_bonus = BonusConfigurationForm(
        data={
            "bonus_sport_func_name": ScoDocSiteConfig.get_bonus_sport_class_name(),
        }
    )
    form_scodoc = ScoDocConfigurationForm(
        data={"enable_entreprises": ScoDocSiteConfig.is_entreprises_enabled()}
    )
    if request.method == "POST" and (
        form_bonus.cancel_bonus.data or form_scodoc.cancel_scodoc.data
    ):  # cancel button
        return redirect(url_for("scodoc.index"))
    if form_bonus.submit_bonus.data and form_bonus.validate():
        if (
            form_bonus.data["bonus_sport_func_name"]
            != ScoDocSiteConfig.get_bonus_sport_class_name()
        ):
            ScoDocSiteConfig.set_bonus_sport_class(
                form_bonus.data["bonus_sport_func_name"]
            )
            app.clear_scodoc_cache()
            flash("""Fonction bonus sport&culture configurée.""")
        else:
            flash("Fonction bonus inchangée.")
        return redirect(url_for("scodoc.index"))
    elif form_scodoc.submit_scodoc.data and form_scodoc.validate():
        if ScoDocSiteConfig.enable_entreprises(
            enabled=form_scodoc.data["enable_entreprises"]
        ):
            flash(
                "Module entreprise "
                + ("activé" if form_scodoc.data["enable_entreprises"] else "désactivé")
            )
        return redirect(url_for("scodoc.index"))

    return render_template(
        "configuration.html",
        form_bonus=form_bonus,
        form_scodoc=form_scodoc,
        scu=scu,
        title="Configuration",
    )
